@@ -10,7 +10,7 @@ Support for Digilent WaveForms devices.
 #
 
 import logging
-from typing import Optional, Tuple
+from typing import Optional, Tuple, Union
 from . import bindings as api
 from .constants import DeviceId, GlobalParameter, TriggerSlope, TriggerSource
 from .exceptions import WaveformsError, DeviceNotFound, DeviceNotOpenError
@@ -27,19 +27,28 @@ from .digital_output import DigitalOutput
 from .protocols import Protocols
 
 
-class DeviceBase():
+class DeviceBase:
     """Base class for Digilent WaveForms devices."""
 
     class _EnumeratedIndex:
         def __init__(self, index):
             self.index = index
 
-    def __init__(self,
-                 configuration=None,
-                 serial_number=None,
-                 device_id=None,
-                 device_type=None,
-                 device_index=None):
+    def __init__(
+        self,
+        configuration: Optional[Union[int, str]] = None,
+        serial_number: Optional[str] = None,
+        device_id: Optional[int] = None,
+        device_type: Optional[int] = None,
+        device_index: Optional[Union[int, _EnumeratedIndex]] = None,
+    ):
+        """
+        :param configuration: Select the active configuration.
+        :param serial_number: Filter devices by serial number.
+        :param device_id: Filter devices by device ID.
+        :param device_type: Filter devices by device type.
+        :param device_index: Filter devices by device index.
+        """
         super().__init__()
 
         self._application = Application()
@@ -57,7 +66,9 @@ class DeviceBase():
         self._auto_reset = True
 
         self._configuration = configuration
-        self._enum_serial_number = DeviceInfo.normalize_serial_number(serial_number)
+        self._enum_serial_number = (
+            DeviceInfo.normalize_serial_number(serial_number) if serial_number else None
+        )
         self._enum_device_id = device_id
         self._enum_device_type = device_type
         self._enum_device_index = device_index
@@ -69,8 +80,8 @@ class DeviceBase():
         self.open()
         return self
 
-    def __exit__(self, _type, _value, _traceback) -> None:
-        del _type, _value, _traceback
+    def __exit__(self, exception_type, exception_value, traceback) -> None:
+        del exception_type, exception_value, traceback
         self.close()
 
     @property
@@ -161,13 +172,14 @@ class DeviceBase():
         return self._device_info.configurations
 
     @property
-    def configuration(self) -> int:
+    def configuration(self) -> Optional[Union[int, str]]:
         """Gets the selected configuration index."""
         return self._configuration
 
     @property
     def auto_configure(self) -> int:
-        """Gets or sets a value indicating to automatically configure the device when parameters are changed."""
+        """Gets or sets a value indicating to automatically configure
+        the device when parameters are changed."""
         self._ensure_handle()
         return api.dwf_device_auto_configure_get(self._hdwf)
 
@@ -215,7 +227,9 @@ class DeviceBase():
         if self._configuration is None:
             self._hdwf = api.dwf_device_open(device_index)
         else:
-            self._hdwf = api.dwf_device_config_open(device_index, self._configuration)
+            configuration = self._get_configuration()
+            self._logger.info('Using configuration %s', configuration)
+            self._hdwf = api.dwf_device_config_open(device_index, configuration)
 
         if self._hdwf is None:
             raise WaveformsError('Failed to open device.')
@@ -284,7 +298,7 @@ class DeviceBase():
 
     def _enum_get_device_index(self) -> Optional[int]:
         if self._enum_device_type:
-            enumfilter = self._enum_device_type
+            enumfilter = api.ENUMFILTER_TYPE | self._enum_device_type
         elif self._enum_device_id:
             enumfilter = self._enum_device_id
         else:
@@ -344,6 +358,40 @@ class DeviceBase():
 
         return module
 
+    def _get_configuration(self):
+        if isinstance(self._configuration, str):
+            if self._configuration == 'scope':
+                if self.id in (DeviceId.ANALOG_DISCOVERY, DeviceId.ANALOG_DISCOVERY2):
+                    return 1
+            if self._configuration == 'pattern':
+                if self.id in (DeviceId.ANALOG_DISCOVERY, DeviceId.ANALOG_DISCOVERY2):
+                    return 2
+            elif self._configuration == 'logic':
+                if self.id in (DeviceId.ANALOG_DISCOVERY, DeviceId.ANALOG_DISCOVERY2):
+                    return 3
+                elif self.id == DeviceId.DIGITAL_DISCOVERY:
+                    return 0
+            elif self._configuration == '1v8':
+                if self.id in (DeviceId.ANALOG_DISCOVERY, DeviceId.ANALOG_DISCOVERY2):
+                    return 4
+            elif self._configuration == 'logic-1v8':
+                if self.id == DeviceId.ANALOG_DISCOVERY2:
+                    return 6
+                elif self.id == DeviceId.DIGITAL_DISCOVERY:
+                    return 0
+            else:
+                raise WaveformsError(
+                    "Invalid configuration: "
+                    "Must be 'scope', 'pattern', 'logic', '1v8', or 'logic-1v8'."
+                )
+
+            raise WaveformsError(
+                f"The device '{self._device_info.name}' does not support "
+                "the configuration '{self._configuration}'."
+            )
+
+        return self._configuration
+
 
 class ElectronicsExplorer(DeviceBase):
     """Digilent Electronics Explorer device."""
@@ -380,7 +428,7 @@ class ElectronicsExplorer(DeviceBase):
                 """Gets the current of the fixed power supply."""
                 return self._device.analog_io[0][2].status
 
-            def setup(self, voltage, enabled=True) -> None:
+            def setup(self, voltage: float, enabled: bool = True) -> None:
                 """Sets up the power supply.
 
                 Parameters
@@ -433,14 +481,16 @@ class ElectronicsExplorer(DeviceBase):
             def current_limit(self, value: bool) -> None:
                 self._device.analog_io[1][2].value = value
 
-            def setup(self, voltage, current_limit=None, enabled=True) -> None:
+            def setup(
+                self, voltage: float, current_limit: float = None, enabled: bool = True
+            ) -> None:
                 """Sets up the positive power supply.
 
                 Parameters
                 ----------
                 voltage : float
                     The output voltage.
-                current_limit : float
+                current_limit : float, optional
                     The power supply current limit in A.
                 enabled : bool, optional
                     If True, then the power supply is enabled (default True).
@@ -490,14 +540,16 @@ class ElectronicsExplorer(DeviceBase):
             def current_limit(self, value: bool) -> None:
                 self._device.analog_io[2][2].value = value
 
-            def setup(self, voltage, current_limit=None, enabled=True) -> None:
+            def setup(
+                self, voltage: float, current_limit: float = None, enabled: bool = True
+            ) -> None:
                 """Sets up the negative power supply.
 
                 Parameters
                 ----------
                 voltage : float
                     The output voltage (must be a negative value).
-                current_limit : float
+                current_limit : float, optional
                     The power supply current limit in A.
                 enabled : bool, optional
                     If True, then the power supply is enabled (default True).
@@ -533,7 +585,7 @@ class ElectronicsExplorer(DeviceBase):
             def voltage(self, value: float) -> None:
                 self._device.analog_io[3][1].value = value
 
-            def setup(self, voltage, enabled=True) -> None:
+            def setup(self, voltage: float, enabled: bool = True) -> None:
                 """Sets up the voltage reference.
 
                 Parameters
@@ -572,7 +624,7 @@ class ElectronicsExplorer(DeviceBase):
             def voltage(self, value: float) -> None:
                 self._device.analog_io[4][1].value = value
 
-            def setup(self, voltage, enabled=True) -> None:
+            def setup(self, voltage: float, enabled: bool = True) -> None:
                 """Sets up the voltage reference.
 
                 Parameters
@@ -660,16 +712,14 @@ class ElectronicsExplorer(DeviceBase):
             """Reads the voltage of voltmeter 4."""
             return self._device.analog_io[8][0].status
 
-    def __init__(self,
-                 configuration=None,
-                 serial_number=None,
-                 device_type=None,
-                 device_index=None):
-        super().__init__(configuration=configuration,
-                         serial_number=serial_number,
-                         device_id=DeviceId.ELECTRONICS_EXPLORER,
-                         device_type=device_type,
-                         device_index=device_index)
+    def __init__(self, configuration=None, serial_number=None, device_type=None, device_index=None):
+        super().__init__(
+            configuration=configuration,
+            serial_number=serial_number,
+            device_id=DeviceId.ELECTRONICS_EXPLORER,
+            device_type=device_type,
+            device_index=device_index,
+        )
         self._supplies = self.Supplies(self)
         self._voltmeters = self.Voltmeters(self)
 
@@ -705,7 +755,7 @@ class AnalogDiscovery(DeviceBase):
             def enabled(self, value: bool) -> None:
                 self._device.analog_io[0][0].value = value
 
-            def setup(self, enabled=True) -> None:
+            def setup(self, enabled: bool = True) -> None:
                 """Sets up the positive power supply.
 
                 Parameters
@@ -731,7 +781,7 @@ class AnalogDiscovery(DeviceBase):
             def enabled(self, value: bool) -> None:
                 self._device.analog_io[1][0].value = value
 
-            def setup(self, enabled=True) -> None:
+            def setup(self, enabled: bool = True) -> None:
                 """Sets up the negative power supply.
 
                 Parameters
@@ -781,16 +831,14 @@ class AnalogDiscovery(DeviceBase):
             """Gets the current taken by the supply regulators."""
             return self._device.analog_io[3][1].status
 
-    def __init__(self,
-                 configuration=None,
-                 serial_number=None,
-                 device_type=None,
-                 device_index=None):
-        super().__init__(configuration=configuration,
-                         serial_number=serial_number,
-                         device_id=DeviceId.ANALOG_DISCOVERY,
-                         device_type=device_type,
-                         device_index=device_index)
+    def __init__(self, configuration=None, serial_number=None, device_type=None, device_index=None):
+        super().__init__(
+            configuration=configuration,
+            serial_number=serial_number,
+            device_id=DeviceId.ANALOG_DISCOVERY,
+            device_type=device_type,
+            device_index=device_index,
+        )
         self._supplies = self.Supplies(self)
 
     @property
@@ -844,7 +892,7 @@ class AnalogDiscovery2(DeviceBase):
             def voltage(self, value: float) -> None:
                 self._device.analog_io[0][1].value = value
 
-            def setup(self, voltage, enabled=True) -> None:
+            def setup(self, voltage: float, enabled: bool = True) -> None:
                 """Sets up the positive power supply.
 
                 Parameters
@@ -883,7 +931,7 @@ class AnalogDiscovery2(DeviceBase):
             def voltage(self, value: float) -> None:
                 self._device.analog_io[1][1].value = value
 
-            def setup(self, voltage, enabled=True) -> None:
+            def setup(self, voltage: float, enabled: bool = True) -> None:
                 """Sets up the negative power supply.
 
                 Parameters
@@ -927,16 +975,14 @@ class AnalogDiscovery2(DeviceBase):
             """Gets the master enable status."""
             return self._device.analog_io.master_enable_status
 
-    def __init__(self,
-                 configuration=None,
-                 serial_number=None,
-                 device_type=None,
-                 device_index=None):
-        super().__init__(configuration=configuration,
-                         serial_number=serial_number,
-                         device_id=DeviceId.ANALOG_DISCOVERY2,
-                         device_type=device_type,
-                         device_index=device_index)
+    def __init__(self, configuration=None, serial_number=None, device_type=None, device_index=None):
+        super().__init__(
+            configuration=configuration,
+            serial_number=serial_number,
+            device_id=DeviceId.ANALOG_DISCOVERY2,
+            device_type=device_type,
+            device_index=device_index,
+        )
         self._supplies = self.Supplies(self)
 
     @property
@@ -991,7 +1037,7 @@ class DigitalDiscovery(DeviceBase):
             def voltage(self, value: float) -> None:
                 self._device.analog_io[0][0].value = value
 
-            def setup(self, voltage) -> None:
+            def setup(self, voltage: float) -> None:
                 """Sets up the power supply.
 
                 Parameters
@@ -1025,16 +1071,14 @@ class DigitalDiscovery(DeviceBase):
             """Gets the master enable status."""
             return self._device.analog_io.master_enable_status
 
-    def __init__(self,
-                 configuration=None,
-                 serial_number=None,
-                 device_type=None,
-                 device_index=None):
-        super().__init__(configuration=configuration,
-                         serial_number=serial_number,
-                         device_id=DeviceId.DIGITAL_DISCOVERY,
-                         device_type=device_type,
-                         device_index=device_index)
+    def __init__(self, configuration=None, serial_number=None, device_type=None, device_index=None):
+        super().__init__(
+            configuration=configuration,
+            serial_number=serial_number,
+            device_id=DeviceId.DIGITAL_DISCOVERY,
+            device_type=device_type,
+            device_index=device_index,
+        )
         self._supplies = self.Supplies(self)
 
     @property
@@ -1128,7 +1172,10 @@ class Device(DeviceBase):
     @staticmethod
     def enumerate(enum_filter=api.ENUMFILTER_ALL) -> tuple:
         """Enumerates all devices."""
-        return tuple(Device(device_index=Device._EnumeratedIndex(i)) for i in range(api.dwf_enum(enum_filter)))
+        return tuple(
+            Device(device_index=Device._EnumeratedIndex(i))
+            for i in range(api.dwf_enum(enum_filter))
+        )
 
     @staticmethod
     def close_all() -> None:
