@@ -9,9 +9,11 @@ Protocols module for Digilent WaveForms devices.
 # SPDX-License-Identifier: MIT
 #
 
+import array
 import ctypes
-from typing import Optional, Tuple
+from typing import Optional, Tuple, Union
 from . import bindings as api
+from .constants import DigitalOutputIdle
 from .helpers import Helpers
 
 
@@ -29,6 +31,7 @@ class Protocols:
             self._data_bits = None
             self._stop_bits = None
             self._parity = None
+            self._inverted = None
 
         @property
         def pin_rx(self) -> Optional[int]:
@@ -90,6 +93,16 @@ class Protocols:
             self._parity = value
             api.dwf_digital_uart_parity_set(self._device.handle, self._map_parity(value))
 
+        @property
+        def inverted(self) -> bool:
+            """Gets or sets the polarity."""
+            return self._inverted
+
+        @inverted.setter
+        def inverted(self, value: bool) -> None:
+            self._inverted = value
+            api.dwf_digital_uart_polarity_set(self._device.handle, value)
+
         def reset(self) -> None:
             """Resets the UART configuration to default value."""
             api.dwf_digital_uart_reset(self._device.handle)
@@ -99,15 +112,18 @@ class Protocols:
             self._data_bits = None
             self._stop_bits = None
             self._parity = None
+            self._inverted = None
 
         def setup(
-                self,
-                pin_rx=None,
-                pin_tx=None,
-                rate=9600,
-                data_bits=8,
-                stop_bits=1,
-                parity='n') -> None:
+            self,
+            pin_rx: Optional[int] = None,
+            pin_tx: Optional[int] = None,
+            rate: int = 9600,
+            data_bits: int = 8,
+            stop_bits: int = 1,
+            parity: str = 'n',
+            inverted: bool = False,
+        ) -> None:
             """Sets up the UART configuration."""
             if pin_rx is not None:
                 self.pin_rx = pin_rx
@@ -121,6 +137,8 @@ class Protocols:
                 self.stop_bits = stop_bits
             if parity is not None:
                 self.parity = parity
+            if inverted is not None:
+                self.inverted = inverted
             if pin_rx is not None:
                 api.dwf_digital_uart_rx(self._device.handle, None, 0)
             if pin_tx is not None:
@@ -129,14 +147,16 @@ class Protocols:
         def read(self, buffer_size=8192) -> Tuple[bytes, int]:
             """Returns the received characters since the last call.
             Returns (rx_buffer, parity)."""
-            c_buffer = (ctypes.c_char * buffer_size)()
-            count, parity = api.dwf_digital_uart_rx(self._device.handle, c_buffer, len(c_buffer))
-            return bytes(c_buffer)[:count], parity
+            rx_buffer8 = (ctypes.c_char * buffer_size)()
+            count, parity = api.dwf_digital_uart_rx(
+                self._device.handle, rx_buffer8, len(rx_buffer8)
+            )
+            return bytes(rx_buffer8)[:count], parity
 
         def write(self, buffer: bytes) -> None:
             """Transmits the specified characters."""
-            c_buffer = (ctypes.c_char * len(buffer)).from_buffer_copy(buffer)
-            api.dwf_digital_uart_tx(self._device.handle, c_buffer, len(c_buffer))
+            tx_buffer8 = (ctypes.c_char * len(buffer)).from_buffer_copy(buffer)
+            api.dwf_digital_uart_tx(self._device.handle, tx_buffer8, len(tx_buffer8))
 
         @staticmethod
         def _map_parity(value) -> int:
@@ -269,33 +289,35 @@ class Protocols:
             self._msb_first = None
 
         def setup(
-                self,
-                pin_clock,
-                mosi_pin,
-                miso_pin=None,
-                pin_select=None,
-                frequency=None,
-                mode=0,
-                msb_first=True) -> None:
+            self,
+            pin_clock: int,
+            pin_mosi: int,
+            pin_miso: Optional[int] = None,
+            pin_select: Optional[int] = None,
+            frequency: Optional[float] = None,
+            mode: int = 0,
+            msb_first: bool = True,
+        ) -> None:
             """Sets up the SPI pin configuration in standard mode."""
             self._dq_mode = 1  # MOSI/MISO
             self.pin_clock = pin_clock
-            if mosi_pin is not None:
-                self.pin_dq0 = mosi_pin
-            if miso_pin is not None:
-                self.pin_dq1 = miso_pin
+            if pin_mosi is not None:
+                self.pin_dq0 = pin_mosi
+            if pin_miso is not None:
+                self.pin_dq1 = pin_miso
             if pin_select is not None:
                 self.pin_select = pin_select
             self._setup_config(frequency, mode, msb_first)
 
         def setup_three_wire(
-                self,
-                pin_clock,
-                pin_siso,
-                pin_select=None,
-                frequency=None,
-                mode=0,
-                msb_first=True) -> None:
+            self,
+            pin_clock: int,
+            pin_siso: int,
+            pin_select: int = None,
+            frequency: float = None,
+            mode: int = 0,
+            msb_first: bool = True,
+        ) -> None:
             """Sets up the SPI pin configuration in Three-wire mode."""
             self._dq_mode = 0  # SISO
             self.pin_clock = pin_clock
@@ -305,14 +327,15 @@ class Protocols:
             self._setup_config(frequency, mode, msb_first)
 
         def setup_dual(
-                self,
-                pin_clock,
-                pin_dq0,
-                pin_dq1,
-                pin_select=None,
-                frequency=None,
-                mode=0,
-                msb_first=True) -> None:
+            self,
+            pin_clock: int,
+            pin_dq0: int,
+            pin_dq1: int,
+            pin_select: int = None,
+            frequency: float = None,
+            mode: int = 0,
+            msb_first: bool = True,
+        ) -> None:
             """Sets up the SPI pin configuration in Dual mode."""
             self._dq_mode = 2  # DUAL
             self.pin_clock = pin_clock
@@ -325,16 +348,17 @@ class Protocols:
             self._setup_config(frequency, mode, msb_first)
 
         def setup_quad(
-                self,
-                pin_clock,
-                pin_dq0,
-                pin_dq1,
-                pin_dq2,
-                pin_dq3,
-                pin_select=None,
-                frequency=None,
-                mode=0,
-                msb_first=True) -> None:
+            self,
+            pin_clock: int,
+            pin_dq0: int,
+            pin_dq1: int,
+            pin_dq2: int,
+            pin_dq3: int,
+            pin_select: int = None,
+            frequency: float = None,
+            mode: int = 0,
+            msb_first: bool = True,
+        ) -> None:
             """Sets up the SPI pin configuration in Quad mode."""
             self._dq_mode = 3  # QUAD
             self.pin_clock = pin_clock
@@ -358,121 +382,140 @@ class Protocols:
             if msb_first is not None:
                 self.msb_first = msb_first
 
-        def select(self, level, pin_select=None) -> None:
+        def set_idle(self, pin: int, idle: DigitalOutputIdle) -> None:
+            """Specifies the DQ signal idle output state.
+            DQ2 and DQ3 may be used for alternative purpose like for write protect
+            (should be driven low) or for hold (should be in high impendance)."""
+            api.dwf_digital_spi_idle_set(self._device.handle, pin, idle)
+
+        def select(self, level: Union[str, int], pin_select: Optional[int] = None) -> None:
             """Control the SPI chip select."""
             if pin_select is None:
                 pin_select = self.pin_select
-            api.dwf_digital_spi_select(self._device.handle, pin_select, self._map_select_level(level))
+            api.dwf_digital_spi_select(
+                self._device.handle, pin_select, self._map_select_level(level)
+            )
 
-        def read_one(self, dq_mode=None, bits_per_word=8) -> None:
+        def read_one(self, dq_mode: Optional[int] = None, bits_per_word: int = 8) -> None:
             """Performs a SPI reception of up to 32 bits."""
             if dq_mode is None:
                 dq_mode = self._dq_mode
             return api.dwf_digital_spi_read_one(self._device.handle, dq_mode, bits_per_word)
 
-        def write_one(self, data: int, dq_mode=None, bits_per_word=8) -> None:
+        def write_one(
+            self, data: int, dq_mode: Optional[int] = None, bits_per_word: int = 8
+        ) -> None:
             """Performs a SPI transmit of up to 32 bits."""
             if dq_mode is None:
                 dq_mode = self._dq_mode
 
             return api.dwf_digital_spi_write_one(self._device.handle, dq_mode, bits_per_word, data)
 
-        def read(self, words_to_receive: int, dq_mode=None, bits_per_word=8):
+        def read(
+            self, words_to_receive: int, dq_mode: Optional[int] = None, bits_per_word: int = 8
+        ) -> Union[bytes, array.array]:
             """Performs a SPI read."""
             if dq_mode is None:
                 dq_mode = self._dq_mode
 
-            c_rx_buffer: ctypes.Array
             if bits_per_word <= 8:
-                c_rx_buffer = (ctypes.c_ubyte * words_to_receive)()
+                rx_buffer8 = (ctypes.c_ubyte * words_to_receive)()
                 api.dwf_digital_spi_read(
-                    self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_rx_buffer, len(c_rx_buffer))
-                return bytes(c_rx_buffer)
-            if bits_per_word <= 16:
-                c_rx_buffer = (ctypes.c_ushort * words_to_receive)()
+                    self._device.handle, dq_mode, bits_per_word, rx_buffer8, len(rx_buffer8)
+                )
+                return bytes(rx_buffer8)
+            elif bits_per_word <= 16:
+                rx_buffer16 = (ctypes.c_ushort * words_to_receive)()
                 api.dwf_digital_spi_read16(
-                    self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_rx_buffer, len(c_rx_buffer))
-                return tuple(c_rx_buffer)
-            if bits_per_word <= 32:
-                c_rx_buffer = (ctypes.c_uint * words_to_receive)()
+                    self._device.handle, dq_mode, bits_per_word, rx_buffer16, len(rx_buffer16)
+                )
+                return array.array('H', rx_buffer16)
+            elif bits_per_word <= 32:
+                rx_buffer32 = (ctypes.c_uint * words_to_receive)()
                 api.dwf_digital_spi_read32(
-                    self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_rx_buffer, len(c_rx_buffer))
-                return tuple(c_rx_buffer)
+                    self._device.handle, dq_mode, bits_per_word, rx_buffer32, len(rx_buffer32)
+                )
+                return array.array('I', rx_buffer32)
+            else:
+                raise ValueError('bits_per_word cannot be higher than 32.')
 
-            raise ValueError('bits_per_word cannot be higher than 32.')
-
-        def write(self, buffer: bytes, dq_mode=None, bits_per_word=8) -> None:
+        def write(
+            self, buffer: bytes, dq_mode: Optional[int] = None, bits_per_word: int = 8
+        ) -> None:
             """Performs a SPI write."""
             if dq_mode is None:
                 dq_mode = self._dq_mode
 
             if bits_per_word <= 8:
-                c_tx_buffer = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
+                tx_buffer8 = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
                 api.dwf_digital_spi_write(
-                    self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_tx_buffer, len(c_tx_buffer))
-            if bits_per_word <= 16:
-                c_tx_buffer = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
+                    self._device.handle, dq_mode, bits_per_word, tx_buffer8, len(tx_buffer8)
+                )
+            elif bits_per_word <= 16:
+                tx_buffer16 = (ctypes.c_ushort * len(buffer)).from_buffer_copy(buffer)
                 api.dwf_digital_spi_write16(
-                    self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_tx_buffer, len(c_tx_buffer))
-            if bits_per_word <= 32:
-                c_tx_buffer = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
+                    self._device.handle, dq_mode, bits_per_word, tx_buffer16, len(tx_buffer16)
+                )
+            elif bits_per_word <= 32:
+                tx_buffer32 = (ctypes.c_uint * len(buffer)).from_buffer_copy(buffer)
                 api.dwf_digital_spi_write32(
-                    self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_tx_buffer, len(c_tx_buffer))
-
-            raise ValueError('bits_per_word cannot be higher than 32.')
+                    self._device.handle, dq_mode, bits_per_word, tx_buffer32, len(tx_buffer32)
+                )
+            else:
+                raise ValueError('bits_per_word cannot be higher than 32.')
 
         def write_read(
-                self,
-                buffer,
-                words_to_receive:
-                int, dq_mode=None,
-                bits_per_word=8):
+            self,
+            buffer,
+            words_to_receive: int,
+            dq_mode: Optional[int] = None,
+            bits_per_word: int = 8,
+        ) -> Union[bytes, array.array]:
             """Performs a SPI write/read."""
             if dq_mode is None:
                 dq_mode = self._dq_mode
 
-            c_tx_buffer: ctypes.Array
-            c_rx_buffer: ctypes.Array
             if bits_per_word <= 8:
-                c_tx_buffer = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
-                c_rx_buffer = (ctypes.c_ubyte * words_to_receive)()
+                tx_buffer8 = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
+                rx_buffer8 = (ctypes.c_ubyte * words_to_receive)()
                 api.dwf_digital_spi_write_read(
                     self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_tx_buffer, len(c_tx_buffer),
-                    c_rx_buffer, len(c_rx_buffer))
-                return bytes(c_rx_buffer)
-            if bits_per_word <= 16:
-                c_tx_buffer = (ctypes.c_ushort * len(buffer)).from_buffer_copy(buffer)
-                c_rx_buffer = (ctypes.c_ushort * words_to_receive)()
+                    dq_mode,
+                    bits_per_word,
+                    tx_buffer8,
+                    len(tx_buffer8),
+                    rx_buffer8,
+                    len(rx_buffer8),
+                )
+                return bytes(rx_buffer8)
+            elif bits_per_word <= 16:
+                tx_buffer16 = (ctypes.c_ushort * len(buffer)).from_buffer_copy(buffer)
+                rx_buffer16 = (ctypes.c_ushort * words_to_receive)()
                 api.dwf_digital_spi_write_read16(
                     self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_tx_buffer, len(c_tx_buffer),
-                    c_rx_buffer, len(c_rx_buffer))
-                return tuple(c_rx_buffer)
-            if bits_per_word <= 32:
-                c_tx_buffer = (ctypes.c_uint * len(buffer)).from_buffer_copy(buffer)
-                c_rx_buffer = (ctypes.c_uint * words_to_receive)()
+                    dq_mode,
+                    bits_per_word,
+                    tx_buffer16,
+                    len(tx_buffer16),
+                    rx_buffer16,
+                    len(rx_buffer16),
+                )
+                return array.array('H', rx_buffer16)
+            elif bits_per_word <= 32:
+                tx_buffer32 = (ctypes.c_uint * len(buffer)).from_buffer_copy(buffer)
+                rx_buffer32 = (ctypes.c_uint * words_to_receive)()
                 api.dwf_digital_spi_write_read32(
                     self._device.handle,
-                    dq_mode, bits_per_word,
-                    c_tx_buffer, len(c_tx_buffer),
-                    c_rx_buffer, len(c_rx_buffer))
-                return tuple(c_rx_buffer)
-            raise ValueError('bits_per_word cannot be higher than 32.')
+                    dq_mode,
+                    bits_per_word,
+                    tx_buffer32,
+                    len(tx_buffer32),
+                    rx_buffer32,
+                    len(rx_buffer32),
+                )
+                return array.array('I', rx_buffer32)
+            else:
+                raise ValueError('bits_per_word cannot be higher than 32.')
 
         @staticmethod
         def _map_select_level(value) -> int:
@@ -494,7 +537,9 @@ class Protocols:
             self._pin_scl = None
             self._pin_sda = None
             self._rate = None
+            self._timeout = None
             self._read_nak = True
+            self._stretch = True
 
         @property
         def pin_scl(self) -> int:
@@ -526,6 +571,16 @@ class Protocols:
             self._rate = value
             api.dwf_digital_i2c_rate_set(self._device.handle, value)
 
+        @property
+        def timeout(self) -> float:
+            """Gets or sets the time-out."""
+            return self._timeout
+
+        @timeout.setter
+        def timeout(self, value: float) -> None:
+            self._timeout = value
+            api.dwf_digital_i2c_timeout_set(self._device.handle, value)
+
         @property
         def read_nak(self) -> bool:
             """Gets or sets a value indicating if the last read byte
@@ -537,20 +592,40 @@ class Protocols:
             self._read_nak = value
             api.dwf_digital_i2c_read_nak_set(self._device.handle, value)
 
+        @property
+        def stretch(self) -> bool:
+            """Enables or disables clock stretching."""
+            return self._stretch
+
+        @stretch.setter
+        def stretch(self, value: bool) -> None:
+            self._stretch = value
+            api.dwf_digital_i2c_stretch_set(self._device.handle, value)
+
         def reset(self) -> None:
             """Resets the I2C configuration to default value."""
             api.dwf_digital_i2c_reset(self._device.handle)
             self._pin_scl = None
             self._pin_sda = None
             self._rate = 100000.0
+            self._timeout = 1.0
             self._read_nak = True
+            self._stretch = True
 
         def clear(self) -> bool:
-            """ Verifies and tries to solve eventual bus lockup.
+            """Verifies and tries to solve eventual bus lockup.
             Returns true, if the bus is free."""
             return bool(api.dwf_digital_i2c_clear(self._device.handle))
 
-        def setup(self, pin_scl, pin_sda, rate=None, read_nak=None) -> None:
+        def setup(
+            self,
+            pin_scl: int,
+            pin_sda: int,
+            rate: Optional[float] = None,
+            timeout: Optional[float] = None,
+            read_nak: Optional[bool] = None,
+            stretch: Optional[bool] = None,
+        ) -> None:
             """Sets up the I2C configuration."""
             if pin_scl is not None:
                 self.pin_scl = pin_scl
@@ -558,8 +633,12 @@ class Protocols:
                 self.pin_sda = pin_sda
             if rate is not None:
                 self.rate = rate
+            if timeout is not None:
+                self.timeout = timeout
             if read_nak is not None:
                 self.read_nak = read_nak
+            if stretch is not None:
+                self.stretch = stretch
 
         def write_one(self, address: int, data: int) -> None:
             """Performs an I2C write of a single byte."""
@@ -568,32 +647,33 @@ class Protocols:
         def read(self, address: int, bytes_to_read: int) -> Tuple[bytes, int]:
             """Performs an I2C read.
             Returns (rx_buffer, nak_index)."""
-            c_rx_buffer = (ctypes.c_ubyte * bytes_to_read)()
+            rx_buffer8 = (ctypes.c_ubyte * bytes_to_read)()
             nak_index = api.dwf_digital_i2c_read(
-                self._device.handle,
-                address,
-                c_rx_buffer, len(c_rx_buffer))
-            return bytes(c_rx_buffer), nak_index
+                self._device.handle, address, rx_buffer8, len(rx_buffer8)
+            )
+            return bytes(rx_buffer8), nak_index
 
         def write(self, address: int, buffer: bytes) -> None:
             """Performs an I2C write."""
-            c_tx_buffer = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
+            tx_buffer8 = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
             return api.dwf_digital_i2c_write(
-                self._device.handle,
-                address,
-                c_tx_buffer, len(c_tx_buffer))
+                self._device.handle, address, tx_buffer8, len(tx_buffer8)
+            )
 
         def write_read(self, address: int, buffer: bytes, bytes_to_read: int) -> Tuple[bytes, int]:
             """Performs an I2C write/read.
             Returns (rx_buffer, nak_index)."""
-            c_tx_buffer = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
-            c_rx_buffer = (ctypes.c_ubyte * bytes_to_read)()
+            tx_buffer8 = (ctypes.c_ubyte * len(buffer)).from_buffer_copy(buffer)
+            rx_buffer8 = (ctypes.c_ubyte * bytes_to_read)()
             nak_index = api.dwf_digital_i2c_write_read(
                 self._device.handle,
                 address,
-                c_tx_buffer, len(c_tx_buffer),
-                c_rx_buffer, len(c_rx_buffer))
-            return bytes(c_rx_buffer), nak_index
+                tx_buffer8,
+                len(tx_buffer8),
+                rx_buffer8,
+                len(rx_buffer8),
+            )
+            return bytes(rx_buffer8), nak_index
 
     class CAN:
         """CAN protocol."""
@@ -653,7 +733,13 @@ class Protocols:
             self._rate = None
             self._inverted = None
 
-        def setup(self, pin_rx=None, pin_tx=None, rate=None, inverted=False) -> None:
+        def setup(
+            self,
+            pin_rx: Optional[int] = None,
+            pin_tx: Optional[int] = None,
+            rate: Optional[float] = None,
+            inverted: bool = False,
+        ) -> None:
             """Sets up the CAN configuration."""
             if pin_rx is not None:
                 self.pin_rx = pin_rx
@@ -668,19 +754,18 @@ class Protocols:
 
         def read(self) -> Tuple[bytes, int, int, int, int]:
             """Returns the received CAN frames since the last call."""
-            c_buffer = (ctypes.c_char * 8)()
+            rx_buffer8 = (ctypes.c_char * 8)()
             frame_id, extended, remote, dlc, status = api.dwf_digital_can_rx(
-                self._device.handle,
-                c_buffer, len(c_buffer))
-            return bytes(c_buffer)[:dlc], frame_id, extended, remote, status
+                self._device.handle, rx_buffer8, len(rx_buffer8)
+            )
+            return bytes(rx_buffer8)[:dlc], frame_id, extended, remote, status
 
         def write(self, frame_id: int, extended: int, remote: int, buffer: bytes) -> None:
             """Performs a CAN transmission."""
-            c_buffer = (ctypes.c_char * len(buffer)).from_buffer_copy(buffer)
+            tx_buffer8 = (ctypes.c_char * len(buffer)).from_buffer_copy(buffer)
             api.dwf_digital_can_tx(
-                self._device.handle,
-                frame_id, extended, remote,
-                len(c_buffer), c_buffer)
+                self._device.handle, frame_id, extended, remote, len(tx_buffer8), tx_buffer8
+            )
 
     def __init__(self, device):
         self._uart = self.Uart(device)
